// file phantom.cpp  

#include <stdio.h>
#include <math.h>

#ifdef MACOSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
extern "C" BOOL initPhantoms(void){return FALSE;}
extern "C" void exitPhantoms(void){;}
extern "C" void UpdateHapticSpace(void){;}
extern "C" void UpdateHapticState(void){;}

#else
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include <HL/hl.h>
#include <HDU/hduMatrix.h>
#include <HDU/hduQuaternion.h>
#include <HDU/hduError.h>

#include <HLU/hlu.h>


extern "C" {  // These need to be defined externally
 extern BOOL   bOffsetPhantoms=FALSE;
 extern double rHapticMaxX,rHapticMaxY,rHapticMaxZ,rHapticMinX,rHapticMinY,rHapticMinZ;
 extern double lHapticMaxX,lHapticMaxY,lHapticMaxZ,lHapticMinX,lHapticMinY,lHapticMinZ;
 extern int    rButton1,rButton2,lButton1,lButton2;
}

extern "C" char phantom1Name[256]="Default PHANToM";
extern "C" char phantom2Name[256]="PHANToM2";

extern "C" double HapticDevicePosition1[3],HapticDevicePosition2[3];
extern "C" double HapticProxyPosition1[3],HapticProxyPosition2[3];
extern "C" double HapticDevRotationMatrix1[4][4],HapticDevRotationMatrix2[4][4];

static BOOL bDraw1=TRUE,bDraw2=TRUE;
static BOOL bProxy1=TRUE,bProxy2=TRUE;   // WHEN TRUE WE SET OUR OWN PROXY POS

static HHD hHD1 = HD_INVALID_HANDLE;
static HHD hHD2 = HD_INVALID_HANDLE;
static HHLRC hHLRC1 = 0;
static HHLRC hHLRC2 = 0;
static HLuint sphereShapeId1;
static HLuint sphereShapeId2;
static HLuint spring1,spring2;

static void exitHandler(void);
static BOOL initHD(HDstring pConfigName, HHD &hHD);
static BOOL initHL(HHD hHD, HHLRC &hHLRC, HLuint &shapeId, HLuint &spring);

static void RenderHapticShape1(HLuint shapeId);
static void RenderHapticShape2(HLuint shapeId);
static void updateWorkspace(int id);
void HLCALLBACK buttonCB1(HLenum event, HLuint object, HLenum thread, 
                       HLcache *cache, void *userdata);
void HLCALLBACK buttonCB2(HLenum event, HLuint object, HLenum thread, 
                       HLcache *cache, void *userdata);

extern "C" BOOL initPhantoms(void){
 BOOL bOnly1=FALSE;
 if(!initHD(phantom1Name, hHD1))return FALSE;   // RIGHT
 if(!initHD(phantom2Name, hHD2))bOnly1=TRUE;    // LEFT 
 // Initialize the contexts and give each one a handle to a device instance
 initHL(hHD1, hHLRC1, sphereShapeId1,spring1);
 hlMakeCurrent(hHLRC1);
 hlBeginFrame();
 hlAddEventCallback(HL_EVENT_1BUTTONDOWN, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB1,(void *)0);
 hlAddEventCallback(HL_EVENT_1BUTTONUP, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB1,(void *)1);
 hlAddEventCallback(HL_EVENT_2BUTTONDOWN, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB1,(void *)10);
 hlAddEventCallback(HL_EVENT_2BUTTONUP, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB1,(void *)11);
 hlEndFrame();
 if(!bOnly1){
 initHL(hHD2, hHLRC2, sphereShapeId2,spring2);
 hlMakeCurrent(hHLRC2);
 hlBeginFrame();
 hlAddEventCallback(HL_EVENT_1BUTTONDOWN, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB2,(void *)3);
 hlAddEventCallback(HL_EVENT_1BUTTONUP, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB2,(void *)4);
 hlAddEventCallback(HL_EVENT_2BUTTONDOWN, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB2,(void *)13);
 hlAddEventCallback(HL_EVENT_2BUTTONUP, HL_OBJECT_ANY, HL_CLIENT_THREAD, 
                       &buttonCB2,(void *)14);
 hlEndFrame();
 }
 hlMakeCurrent(NULL);
 return TRUE;
}

extern "C" void exitPhantoms(void){
  exitHandler();
  //MessageBox(NULL,"Phantoms freed","OK",MB_OK);
}

extern "C" void UpdateHapticSpace(void){
 if (hHLRC1 != NULL){hlMakeCurrent(hHLRC1); updateWorkspace(1);}  //right
 //else MessageBox(NULL,"Bad context Haptic 1",NULL,MB_OK);
 hlMakeCurrent(NULL);
 if (hHLRC2 != NULL){hlMakeCurrent(hHLRC2); updateWorkspace(2);}  // left
 //else MessageBox(NULL,"Bad context Haptic 2",NULL,MB_OK);
 hlMakeCurrent(NULL);
}

extern "C" void UpdateHapticState(void){
 HLdouble xform[16];
 //if((GetKeyState(VK_F3)&0x8000))return;  //To hold steady
 hlMakeCurrent(hHLRC1);
 hlGetDoublev(HL_DEVICE_POSITION, HapticDevicePosition1);  
 hlGetDoublev(HL_PROXY_POSITION, HapticProxyPosition1);
 hlGetDoublev(HL_DEVICE_TRANSFORM,xform);
 HapticDevRotationMatrix1[0][0]=xform[0]; HapticDevRotationMatrix1[0][1]=xform[4]; HapticDevRotationMatrix1[0][2]=xform[ 8]; HapticDevRotationMatrix1[0][3]=0.0;
 HapticDevRotationMatrix1[1][0]=xform[1]; HapticDevRotationMatrix1[1][1]=xform[5]; HapticDevRotationMatrix1[1][2]=xform[ 9]; HapticDevRotationMatrix1[1][3]=0.0;
 HapticDevRotationMatrix1[2][0]=xform[2]; HapticDevRotationMatrix1[2][1]=xform[6]; HapticDevRotationMatrix1[2][2]=xform[10]; HapticDevRotationMatrix1[2][3]=0.0;
 HapticDevRotationMatrix1[3][0]=0.0;      HapticDevRotationMatrix1[3][1]=0.0;      HapticDevRotationMatrix1[3][2]=0.0;       HapticDevRotationMatrix1[3][3]=1.0;      
 RenderHapticShape1(sphereShapeId1);
 hlMakeCurrent(hHLRC2);
 if(hHLRC2 == NULL){
   int i,j;
   for(i=0;i<3;i++){HapticProxyPosition2[i]=HapticProxyPosition1[i]; HapticDevicePosition2[i]=HapticDevicePosition1[i];}
   for(i=0;i<4;i++)for(j=0;j<4;j++)HapticDevRotationMatrix2[i][j]=HapticDevRotationMatrix1[i][j];
   for(i=0;i<4;i++)HapticDevRotationMatrix2[i][i]=1.0;
   return; // only one phantom
 }
 hlGetDoublev(HL_DEVICE_POSITION, HapticDevicePosition2); 
 hlGetDoublev(HL_PROXY_POSITION, HapticProxyPosition2);
 hlGetDoublev(HL_DEVICE_TRANSFORM,xform);     //hlGetDoublev(HL_DEVICE_ROTATION,proxyqua2);
 HapticDevRotationMatrix2[0][0]=xform[0]; HapticDevRotationMatrix2[0][1]=xform[4]; HapticDevRotationMatrix2[0][2]=xform[ 8]; HapticDevRotationMatrix2[0][3]=0.0;
 HapticDevRotationMatrix2[1][0]=xform[1]; HapticDevRotationMatrix2[1][1]=xform[5]; HapticDevRotationMatrix2[1][2]=xform[ 9]; HapticDevRotationMatrix2[1][3]=0.0;
 HapticDevRotationMatrix2[2][0]=xform[2]; HapticDevRotationMatrix2[2][1]=xform[6]; HapticDevRotationMatrix2[2][2]=xform[10]; HapticDevRotationMatrix2[2][3]=0.0;
 HapticDevRotationMatrix2[3][0]=0.0;      HapticDevRotationMatrix2[3][1]=0.0;      HapticDevRotationMatrix2[3][2]=0.0;       HapticDevRotationMatrix2[3][3]=1.0;      
 RenderHapticShape2(sphereShapeId2);
 hlMakeCurrent(NULL);
 return;
}

extern "C" void SetPhantomForce(int type, int joyID){   // joyID = 1 for first joystick (right) May need to swap
 HDdouble anchor[3];
 if(type == 0){ // NOT USED
 }
 else if(type == 1){ // NOT USED
 }
 else if(type == 2){ // start   effect; 1 = friction 2 = spring
  if(joyID == 1 || joyID == 2){
     if(joyID == 1)bDraw1=FALSE;
     else          bDraw2=FALSE;
     if(joyID == 1)hlMakeCurrent(hHLRC1);
     else          hlMakeCurrent(hHLRC2);
     hlBeginFrame();
     hlPushMatrix();
     hlLoadIdentity();
     //hlEffectd(HL_EFFECT_PROPERTY_GAIN, 0.5);       // strong
     //hlEffectd(HL_EFFECT_PROPERTY_MAGNITUDE, 4.0);  // strong
     hlEffectd(HL_EFFECT_PROPERTY_GAIN, 0.25);
     hlEffectd(HL_EFFECT_PROPERTY_MAGNITUDE, 0.6);  //1.0 (medium)  0.5  (weak)
     hlGetDoublev(HL_DEVICE_POSITION, anchor);
     hlEffectdv(HL_EFFECT_PROPERTY_POSITION, anchor);
     if(joyID == 1)hlStartEffect(HL_EFFECT_SPRING, spring1);
     else          hlStartEffect(HL_EFFECT_SPRING, spring2);
     hlPopMatrix();
     hlEndFrame();
     hlMakeCurrent(NULL);
   }
 }
 else if(type == 3){ // stop  effect
   if(joyID == 1 || joyID == 2){
     if(joyID == 1)hlMakeCurrent(hHLRC1);
     else          hlMakeCurrent(hHLRC2);
     hlBeginFrame();
     if(joyID == 1)hlStopEffect(spring1);
     else          hlStopEffect(spring2);
     hlEndFrame();
     hlMakeCurrent(NULL);
     if(joyID == 1)bDraw1=TRUE;
     else          bDraw2=TRUE;
   }
 }
 else if(type == 4){ // joysticks have struck eack other - push apart
   static const HDdouble direction1[3] = { 0,  1,  1 };
   static const HDdouble direction2[3] = { 0, -1, -1 };
   static const HDdouble duration = 100; /* ms */
   static const HDdouble mag=0.5;
   hlMakeCurrent(hHLRC1);
   hlBeginFrame();
   hlEffectd(HL_EFFECT_PROPERTY_DURATION, duration);
   hlEffectd(HL_EFFECT_PROPERTY_MAGNITUDE, mag);
   hlEffectdv(HL_EFFECT_PROPERTY_DIRECTION, direction1);
   hlTriggerEffect(HL_EFFECT_CONSTANT);
   hlEndFrame();
   hlMakeCurrent(NULL);
   Sleep(102);
   hlMakeCurrent(hHLRC2);
   hlBeginFrame();
   hlEffectd(HL_EFFECT_PROPERTY_DURATION, duration);
   hlEffectd(HL_EFFECT_PROPERTY_MAGNITUDE, mag);
   hlEffectdv(HL_EFFECT_PROPERTY_DIRECTION, direction2);
   hlTriggerEffect(HL_EFFECT_CONSTANT);
   hlEndFrame();
   hlMakeCurrent(NULL);
   Sleep(102);
 }
 else if(type == 5){ // start  send poke - when hits surface - not used
 }
}

extern "C" void TogglePhantomProxy(int id, BOOL on){
 static BOOL last_state1=FALSE,last_state2;
 if(id == 1){ if(on == last_state1)return; last_state1=on;}
 if(id == 2){ if(on == last_state2)return; last_state2=on;}
 if(id == 1)hlMakeCurrent(hHLRC1);  // right
 else       hlMakeCurrent(hHLRC2);  // left
 hlBeginFrame();
 if(!on){    // Don't use proxy rendering
   hlEnable(HL_PROXY_RESOLUTION); 
   if(id == 1)bProxy1=TRUE; else bProxy2=TRUE;
 }
 else{
   hlDisable(HL_PROXY_RESOLUTION);
   hlProxyf(HL_STIFFNESS,(HLfloat)0.18);   // 0.25
   hlProxyf(HL_DAMPING,(HLfloat)0.2);
   if(id == 1){hlProxydv(HL_PROXY_POSITION,HapticDevicePosition1); bProxy1=FALSE;}
   else       {hlProxydv(HL_PROXY_POSITION,HapticDevicePosition2); bProxy2=FALSE;}
 }
 hlEndFrame();
 hlMakeCurrent(NULL);
}

extern "C" void OffsetPhantomProxy(int id, double dalpha, double dphi, double dtheta){
 if(1){//if(fabs(dtheta) > 5.0 || fabs(dphi) > 5.0){
   if(id == 1)hlMakeCurrent(hHLRC1);  // right 
   else       hlMakeCurrent(hHLRC2);  // left
   hlBeginFrame();
   if(id == 1 && bProxy1 == FALSE){
     HapticProxyPosition1[0] += 1.0;
     HapticProxyPosition1[1] += 0.0*dtheta;
     hlProxydv(HL_PROXY_POSITION,HapticProxyPosition1);
   }
   else if(id == 2 && bProxy2 == FALSE){
   }
   hlEndFrame();
   hlMakeCurrent(NULL);
 }
}

static BOOL initHD(HDstring pConfigName, HHD &hHD){
 HDErrorInfo error;
 hHD = hdInitDevice(pConfigName);
 if (HD_DEVICE_ERROR(error = hdGetError()))    {
   //MessageBox(NULL,"ERROR Cannot initialise Phantom",pConfigName,MB_OK); 
   return FALSE;
 }
 //MessageBox(NULL,pConfigName,"Phantom OK",MB_OK);
 return TRUE;
}

static BOOL initHL(HHD hHD, HHLRC &hHLRC, HLuint &shapeId, HLuint &spring){    
 hHLRC = hlCreateContext(hHD);
 hlMakeCurrent(hHLRC);
 // Enable optimization of the viewing parameters when rendering
 // geometry for OpenHaptics
 hlEnable(HL_HAPTIC_CAMERA_VIEW);
 hlEnable(HL_ADAPTIVE_VIEWPORT);
 // Specify front face touchability only
 hlTouchableFace(HL_FRONT);
 // generate id's for the three shapes
 shapeId = hlGenShapes(1);
 spring = hlGenEffects(1);
 if(!hlIsEffect(spring))MessageBox(NULL,"ERROR bad spring effect",NULL,MB_OK);
 return TRUE;
}

static void exitHandler() {
    if (hHLRC1 != NULL)    {
      hlDeleteShapes(sphereShapeId1, 1);
      hlMakeCurrent(hHLRC1);
      hlBeginFrame();
      hlStopEffect(spring1);
      hlEndFrame();
      hlDeleteEffects(spring1, 1);
    }
    if (hHLRC2 != NULL)    {
      hlDeleteShapes(sphereShapeId2, 1);
      hlMakeCurrent(hHLRC2);
      hlBeginFrame();
      hlStopEffect(spring2);
      hlEndFrame();
      hlDeleteEffects(spring2, 1);
    }
    hlMakeCurrent(NULL);
    if (hHLRC1 != NULL)    {
        hlDeleteContext(hHLRC1);
    }
    if (hHLRC2 != NULL)    {
        hlDeleteContext(hHLRC2);
    }
    if (hHD1 != HD_INVALID_HANDLE)    {
        hdDisableDevice(hHD1);
    }
    if (hHD2 != HD_INVALID_HANDLE)    {
        hdDisableDevice(hHD2);
    }
}

static void HLCALLBACK buttonCB1(HLenum event, HLuint object, HLenum thread, 
                         HLcache *cache, void *userdata){
 HLuint s = ((HLuint) userdata); 
 hlMakeCurrent(hHLRC1);
 if (event == HL_EVENT_1BUTTONDOWN){
   rButton1=1;
 }
 else if (event == HL_EVENT_1BUTTONUP){
   rButton1=0;
 }
 else if (event == HL_EVENT_2BUTTONDOWN){
   rButton2=1;
 }
 else if (event == HL_EVENT_2BUTTONUP){
   rButton2=0;
 }
 hlMakeCurrent(NULL);
}

static void HLCALLBACK buttonCB2(HLenum event, HLuint object, HLenum thread, 
                         HLcache *cache, void *userdata){
 HLuint s = ((HLuint) userdata); 
 hlMakeCurrent(hHLRC2);
 if (event == HL_EVENT_1BUTTONDOWN){
   lButton1=1;
 }
 else if (event == HL_EVENT_1BUTTONUP){
   lButton1=0;
 }
 else if (event == HL_EVENT_2BUTTONDOWN){
   lButton2=1;
}
 else if (event == HL_EVENT_2BUTTONUP){
   lButton1=0;
 }
 hlMakeCurrent(NULL);
}

static void updateWorkspace(int id ){  // 1=right 2=left
 GLdouble modelview[16];
 GLdouble projection[16];
 GLint viewport[4];
 HLdouble minWorkBox[3]; 
 HLdouble maxWorkBox[3]; // (OGL coord system)
 glPushMatrix();
 glLoadIdentity();
 glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
 glGetDoublev(GL_PROJECTION_MATRIX, projection);
 glGetIntegerv(GL_VIEWPORT, viewport);
 glPopMatrix();
 hlMatrixMode(HL_TOUCHWORKSPACE);

 hlLoadIdentity();
 /* fit haptic workspace to view volume  - trial and error */
 { // when pivot is within the claspers
   if(id == 1){ // right
     minWorkBox[0]=rHapticMinX; minWorkBox[1]=rHapticMinY; minWorkBox[2]=rHapticMinZ; 
     maxWorkBox[0]=rHapticMaxX; maxWorkBox[1]=rHapticMaxY; maxWorkBox[2]=rHapticMaxZ;
   }
   else{ // left
     minWorkBox[0]=lHapticMinX; minWorkBox[1]=lHapticMinY; minWorkBox[2]=lHapticMinZ; 
     maxWorkBox[0]=lHapticMaxX; maxWorkBox[1]=lHapticMaxY; maxWorkBox[2]=lHapticMaxZ;
   }
 }
 if(bOffsetPhantoms){  // x, y, z OpenGL coord system
   // To allow for the tool pivot being moved back along the shaft
   minWorkBox[2] += 8.0;
   maxWorkBox[2] -= 4.0;
 }

 hluFitWorkspaceBox(modelview,minWorkBox,maxWorkBox);
 //printf("Update Workspace\n");
 //printf("(%f %f)\n (%f %f)\n (%f %f)\n",minWorkBox[0],maxWorkBox[0],minWorkBox[1],maxWorkBox[1],minWorkBox[2],maxWorkBox[2]);
}

HLfloat surface_stiffness1=0.4f;
HLfloat surface_Dfriction1=0.2f;
HLfloat surface_Sfriction1=0.2f;

HLfloat surface_stiffness2=0.4f;
HLfloat surface_Dfriction2=0.2f;
HLfloat surface_Sfriction2=0.2f;

static void RenderHapticShape1(HLuint shapeId){  // actually render the haptic surfaces
  glClear(GL_DEPTH_BUFFER_BIT);         // right
  hlBeginFrame();
  //    hlBeginShape(HL_SHAPE_FEEDBACK_BUFFER, shapeId);
  hlBeginShape(HL_SHAPE_DEPTH_BUFFER, shapeId);
  if(bDraw1){
    hlMaterialf(HL_FRONT_AND_BACK, HL_STIFFNESS,surface_stiffness1);
    //hlMaterialf(HL_FRONT, HL_DAMPING,surface_damping1);
    hlMaterialf(HL_FRONT, HL_STATIC_FRICTION,surface_Sfriction1);
    hlMaterialf(HL_FRONT, HL_DYNAMIC_FRICTION,surface_Dfriction1);
    //drawHapticScene(1); 
  }
  hlEndShape();
  hlCheckEvents();
  hlEndFrame();
}

static void RenderHapticShape2(HLuint shapeId){  // left tool
  glClear(GL_DEPTH_BUFFER_BIT);   
  hlBeginFrame();
  //    hlBeginShape(HL_SHAPE_FEEDBACK_BUFFER, shapeId);
  hlBeginShape(HL_SHAPE_DEPTH_BUFFER, shapeId);
  if(bDraw2){
    hlMaterialf(HL_FRONT_AND_BACK, HL_STIFFNESS,surface_stiffness2);
    //hlMaterialf(HL_FRONT, HL_DAMPING,surface_damping2);
    hlMaterialf(HL_FRONT, HL_STATIC_FRICTION,surface_Sfriction2);
    hlMaterialf(HL_FRONT, HL_DYNAMIC_FRICTION,surface_Dfriction2);
    //drawHapticScene(2); 
  }
  hlEndShape();
  hlCheckEvents();
  hlEndFrame();
}

#endif   // MAC OS
