// All matrix operation tke use the conventional transformation notation
// i.e.  the transformation PRE-multiplies the position vector.  p' = [a]p
// 

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>

#ifdef MACOSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "eulerangles.h"

#ifndef min
#define min(a,b)  ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
#define max(a,b)  ( ((a) > (b)) ? (a) : (b) )
#endif
#define BOUND(x,min,max) ((x) < (min) ? (min) : ((x) > (max) ? (max) : (x)))
#define ALIGNULONG(i)    ((i+3)/4*4)
#define VECCOPY(a,b)    { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; }
#define VECCOPY4(a,b)   { b[0] = a[0]; b[1] = a[1]; b[2] = a[2]; b[3] = a[3];}
#define INVERT(a)       { a[0] = -a[0]; a[1] = -a[1]; a[2] = -a[2]; }
#define VECSUB(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2];}
#define VECSUM(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2];}
#define VECSUB4(a,b,c)   { c[0]=a[0]-b[0]; c[1]=a[1]-b[1]; c[2]=a[2]-b[2]; c[3]=a[3]-b[3];}
#define VECSUM4(a,b,c)   { c[0]=a[0]+b[0]; c[1]=a[1]+b[1]; c[2]=a[2]+b[2]; c[3]=a[3]+b[3];}
#define VECSCALE(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2];}
#define VECSCALE4(a,b,c) { c[0]=(a)*b[0]; c[1]=(a)*b[1]; c[2]=(a)*b[2]; c[3]=b[3];}
#define VECSUMSCALE(s,a,b,c)  \
                        { c[0]=s*(a[0]+b[0]); c[1]=s*(a[1]+b[1]); c[2]=s*(a[2]+b[2]);}
#define DOT(a,b)        ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define DOT4(a,b)       ( (a[0]*b[0]) + (a[1]*b[1]) + (a[2]*b[2]) )
#define CROSS(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                      }
#define CROSS4(v1,v2,r)  { \
                          r[0] = (v1[1]*v2[2]) - (v2[1]*v1[2]);  \
                          r[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);  \
                          r[2] = (v1[0]*v2[1]) - (v2[0]*v1[1]);  \
                          r[3] = 1.0; \
                      }
#define LENGTH(a)       sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2])
#define LENGTH2(a)      (a[0]*a[0]+a[1]*a[1]+a[2]*a[2])
#define DEGRAD 3.1415926/180.0
#define RADDEG 180.0/3.1415926

#define BASEHEIGHT  31.5
#define UPPERARM    45.0
#define LOWERARM    50.0
#define HAND         8.0
#define TOFFSET     15.0
#define SEPARATION 100.0

// These are in the ROBOT coord system  X to front, Y to side, Z up  (dimensions CMS)
#if 0
// use these for the FULL workspace
#define BASE_X -55 
#define BASE_Y   0
#define BASE_Z  70
#define SIZE_X  55
#define SIZE_Y  50
#define SIZE_Z  80
#else
// use these for the SMALL workspace
// a bit further forward and a bit bigger
#define BASE_X -55    
#define BASE_Y   0
#define BASE_Z  45
#define SIZE_X   8
#define SIZE_Y   8
#define SIZE_Z   8
#endif

typedef float  vector[3];
typedef long   ivector[3];
typedef float  normal[3];
typedef float  quaternion[4];

// These are set/returned from the Phantom using OGL coords  i.e.  the Z axis is in/out  
// y is up/down  and X is left/right
// The robot is also constructed using OGL coords - its base is at (0,0,0) but it grows upwards 
// so appears to lie on its side. BUT we change the camera so it it appears vertical - we will
// thus need to swizzle Phantom data so that an up down movement goes along the Z axis an inout
// movement goes along the Y axis and a left/right along x. (We will also have to arrange that 
// this is RELATIVE to the robot and not our viewpoint!!!!! which starts off looking at the front of the robot.

//double rHapticMaxX=50,rHapticMaxY=150,rHapticMaxZ=20,rHapticMinX=-50,rHapticMinY=-10,rHapticMinZ=-90;
//double lHapticMaxX=50,lHapticMaxY=150,lHapticMaxZ=20,lHapticMinX=-50,lHapticMinY=-10,lHapticMinZ=-90;
double rHapticMaxX=BASE_Y+SIZE_Y,
       rHapticMaxY=BASE_Z+SIZE_Z,
       rHapticMaxZ=BASE_X+SIZE_X,
       rHapticMinX=BASE_Y-SIZE_Y,
       rHapticMinY=BASE_Z-SIZE_Z,
       rHapticMinZ=BASE_X-SIZE_X;
double lHapticMaxX=BASE_Y+SIZE_Y,
       lHapticMaxY=BASE_Z+SIZE_Z,
       lHapticMaxZ=BASE_X+SIZE_X,
       lHapticMinX=BASE_Y-SIZE_Y,
       lHapticMinY=BASE_Z-SIZE_Z,
       lHapticMinZ=BASE_X-SIZE_X;

double HapticDevicePosition1[3],HapticDevicePosition2[3];
double HapticProxyPosition1[3],HapticProxyPosition2[3];
double HapticDevRotationMatrix1[4][4],HapticDevRotationMatrix2[4][4];
int    rButton1=0,rButton2=0,lButton1=0,lButton2=0;

extern BOOL initPhantoms(void);
extern void exitPhantoms(void);
extern void UpdateHapticSpace(void);
extern void UpdateHapticState(void);
extern void SetPhantomForce(int type, int joyID);
int  bPhantoms=0;

float robo_nxyz[3],robo_oxyz[3],robo_axyz[3],robo_pxyz[3]; 
float robo2_nxyz[3],robo2_oxyz[3],robo2_axyz[3],robo2_pxyz[3]; 
EulerAngles Robo1_E,Robo2_E;

GLfloat rs1=0.0,rs2=0.0,rs3=0.0,re1=0.0,re2=0.0,rw1=0.0,rw2=0.0;
GLfloat qs1=0.0,qs2=0.0,qs3=0.0,qe1=0.0,qe2=0.0,qw1=0.0,qw2=0.0;

GLfloat hand_pan=0.0,hand_tilt=0.0,hand_twist=0.0;
GLfloat ccd_rs1=0.0,ccd_rs2=0.0,ccd_rs3=0.0,ccd_re1=0.0,ccd_re2=0.0,ccd_rw1=0.0,ccd_rw2=0.0;

GLfloat rs1a[10],rs2a[10],rs3a[10],re1a[10],re2a[10],rw1a[10],rw2a[10];

GLfloat baseHeight=BASEHEIGHT
       ,upperArm=UPPERARM
       ,lowerArm=LOWERARM
       ,Hand=HAND;
int viewdir=1,modelVis=1,axesVis=0,ccdVis=0,get_transform=0,ccdRLT=1,interactive=0,continuous=0;
GLfloat m_camera[4][4],
        m_endpoint[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},
        m_wristpoint[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},
        m_lowerarm[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},
        m_upperarm[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},
        p_endpoint[3]={20.0,0.0,0.0},
        p_upperarm[3],
        p_lowerarm[3],
        p_wrist[3]={0,0,0},
        p_target[3]={42,0,50},     // CCD algorithm WRIST target
        p_target2[3]={42,0,50},    // and for the second robot
        p_hand[3]={50,0,50},
        p_hand2[3]={50,0,50};      // hand coodinates

float   c_upperarm[4]={0,0,BASEHEIGHT,1},            // these are for the CCD algorithm
        c_lowerarm[4]={0,0,BASEHEIGHT+UPPERARM,1},
        c_wrist[4]={0,0,BASEHEIGHT+UPPERARM+LOWERARM,1},
        c_hand[4]={0,0,BASEHEIGHT+UPPERARM+LOWERARM+HAND,1},
        c_target[4]={TOFFSET,0,BASEHEIGHT+UPPERARM+LOWERARM,1},  // this is a point for the hand to point at
        II[4][4]={1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},   // Identity 
        B[4][4],
        U[4][4],
        L[4][4],
        H[4][4],
        BU[4][4],
        BUL[4][4],
        BULH[4][4],
        ToolO[4][4]={1,0,0,0, 0,1,0,0,  0,0,1,0,  0,0,0,1},
        ToolO2[4][4]={1,0,0,0, 0,1,0,0,  0,0,1,0,  0,0,0,1};

static int  InvertMatrix(float *A, long n, float *B);
static void InvertMatrixRotation(float T[4][4], float t[4][4], int zero_translation);
static void InvertMatrixRotationT(float T[4][4], float t[4][4], int not_used);
static void TransposeMatrix16(GLfloat m[16]);
static void m4by4(GLfloat t1[4][4], GLfloat t2[4][4], GLfloat tr[4][4]);
static void m16to4(GLfloat m[16], GLfloat v[4][4]);
static void m4to16(GLfloat v[4][4], GLfloat m[16]);
static void m4by1(GLfloat t[4][4], GLfloat m[4],  GLfloat r[4]);
static void rotz(float tr[4][4], float ang);
static void rotx(float tr[4][4], float ang);
static void roty(float tr[4][4], float ang);
static void tram(float t[4][4], float dx, float dy, float dz);
static void c4to4(float tin[4][4], float tout[4][4]);
static void ConvertHLtoLocal(float mgl[4][4], float mfx[4][4]);
static void null_transform(float t[4][4]);
static void ExtractMatrixRotation(float T[4][4], float t[4][4]);
static void SetMatrixRotation(float T[4][4], float r[4][4], float t[4][4]);
static void RunCCD(int prnt, int id);
static void MatrixToEulerAngles(float m[4][4], float *f, float *t, float *a );
static void normalise(float v[4]);
static void RotateVector(float ct, float st, float r[4], float n[4]);
static void SetHandDirection(int id);
static void UpdateFromPhantoms(void);
static void GetAnglesFromFile(void);
static void verifyEulerConversion(HMatrix, HMatrix);


static GLdouble er=200.0,theta=0.0,phi=0.0;
static float angx,angy,angz;

void UpdateRobotServerData(void){
 int i;
 float x[4]={1,0,0,1},y[4]={0,1,0,1},z[4]={0,0,1,1}; // These may need to change for correct robot operation
 float n[4],o[4],a[4];
 m4by1(ToolO,x,n);
 m4by1(ToolO,y,o);
 m4by1(ToolO,z,a);
 for(i=0;i<3;i++){
   //robo_pxyz[i]=c_wrist[i];
   robo_pxyz[i]=p_hand[i];  // it is the hand we are positioning
   robo_nxyz[i]=n[i];
   robo_oxyz[i]=o[i];
   robo_axyz[i]=a[i];
 }
 m4by1(ToolO2,x,n);
 m4by1(ToolO2,y,o);
 m4by1(ToolO2,z,a);
 for(i=0;i<3;i++){
   //robo_pxyz[i]=c_wrist[i];
   robo2_pxyz[i]=p_hand2[i];  //
   robo2_nxyz[i]=n[i];
   robo2_oxyz[i]=o[i];
   robo2_axyz[i]=a[i];
 }
 verifyEulerConversion(ToolO,ToolO2);
}

void drawDisk(float r) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluDisk(pObj,0.0f,r,30,3);
  gluDeleteQuadric(pObj);
  glPopMatrix();
}

void drawSphere(float r) {
  GLUquadricObj* pObj = gluNewQuadric();
  gluSphere(pObj, r, 30, 30);
  gluDeleteQuadric(pObj);
}

void drawCylinder(float r, float h) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluCylinder(pObj,r,r,h,30,10);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,30,3);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,30,3);
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawWireCylinder(float r, float h, float b) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  gluQuadricDrawStyle(pObj,GLU_LINE);
  //  gluQuadricDrawStyle(pObj,GLU_SILHOUETTE);
  glLineWidth(0.25);
  glTranslatef(0.0f,0.0f,b);
  gluCylinder(pObj,r,r,h,15,5);
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,15,1);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,15,1);
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawCylinderOffset(float r, float h, float b) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  glColorMaterial(GL_FRONT,GL_DIFFUSE);  
  glEnable(GL_COLOR_MATERIAL);
  glColor3f(1.0,0.8,0.8);
  glTranslatef(0.0f,0.0f,b);
  gluCylinder(pObj,r,r,h,15,5);
  glTranslatef(0.0f,0.0f,h);
  gluDisk(pObj,0.0f,r,15,1);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,15,1);
  glPopMatrix();
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_COLOR_MATERIAL);
  gluDeleteQuadric(pObj);
}

void drawCylinderH(float r, float h) {
  GLUquadricObj* pObj = gluNewQuadric();
  glPushMatrix();
  glRotatef(90.0,1.0,0.0,0.0);
  glPushMatrix();
  glTranslatef(0.0f,0.0f,-h/2.0);
  gluCylinder(pObj,r,r,h,30,10);
  glPopMatrix();
  glPushMatrix();
  glTranslatef(0.0f,0.0f,h/2.0);
  gluDisk(pObj,0.0f,r,30,3);
  glTranslatef(0.0f,0.0f,-h);
  gluQuadricOrientation(pObj,GLU_INSIDE);
  gluDisk(pObj,0.0f,r,30,3);
  glPopMatrix();
  glPopMatrix();
  gluDeleteQuadric(pObj);
}

void drawAxes(float l){
   glDisable(GL_LIGHTING);
   glColorMaterial(GL_FRONT,GL_DIFFUSE);  
   glEnable(GL_COLOR_MATERIAL);
   glLineWidth(4.0);
   glColor3f(1.0,0.0,0.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(l,  0.0,0.0);
   glColor3f(0.0,1.0,0.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(0.0,  l,0.0);
   glColor3f(0.0,0.0,1.0);
   glBegin(GL_LINES);
   glVertex3f(0.0,0.0,0.0);
   glVertex3f(0.0,0.0, l);
   glEnd();
   glColor3f(1.0,1.0,1.0);
   glDisable(GL_COLOR_MATERIAL);
   glEnable(GL_LIGHTING);
}

void drawTargetAndProxy(void){
  {
  float mm[4][4],mmm[16],mmi[4][4],mmr[4][4],mmu[4][4];
  glPushMatrix();     // draw AXES for the end point axes
  glLoadIdentity(); 

  //ExtractMatrixRotation(m_endpoint,mmr);  // These steps undo the rotation part of the end
  //InvertMatrixRotation(mmr,mmi,1);        // effector so that it aligns with global 
  //m4by4(mmi,mmr,mmu);                     // frame of reference (at the base of the robot.)
  //SetMatrixRotation(m_endpoint,mmu,mmr);  // ..
  //m4by4(m_camera,mmr,mm);                 // .. They could replace the next line
  m4by4(m_camera,m_endpoint,mm);  // compose position and camera view transform

  m4to16(mm,mmm);
  TransposeMatrix16(mmm);
  glLoadMatrixf(mmm);
  drawAxes(15.0);
  glPopMatrix();      // end of endpoint axes
  }
  glPushMatrix();     // draw proxy of last endpoint reached
  glTranslatef(p_endpoint[0],p_endpoint[1],p_endpoint[2]);
  glEnable(GL_COLOR_MATERIAL);
  glColor3f(0.5,1.0,0.5);
  drawSphere(1.2);
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_COLOR_MATERIAL);
  glPopMatrix();      //  end proxy endpoint

  glEnable(GL_COLOR_MATERIAL);
  glColor3f(1.0,1.0,0.5);
  glPushMatrix();     // target for CCD algorithm to place wrist at 
  glTranslatef(p_target[0],p_target[1],p_target[2]);
  drawSphere(1.2);
  glPopMatrix();    
  glPushMatrix();     // target for CCD algorithm to place wrist at 
  glTranslatef(p_target2[0],p_target2[1]+SEPARATION,p_target2[2]);
  drawSphere(1.2);
  glPopMatrix();    
  if(rButton1 || rButton2){
    glPushMatrix();    
    glColor3f(0.7,0.0,0.0); // target for ccd_hand to point at (just to show point is working)
    glTranslatef(c_target[0],c_target[1],c_target[2]);
    drawSphere(1.0);
    glPopMatrix();
  }
  if(bPhantoms > 0){ // Draw the phantom device positions
    glPushMatrix();    
    glColor3f(1.0,0.0,1.0); 
    glTranslatef(-HapticDevicePosition1[2],-HapticDevicePosition1[0],HapticDevicePosition1[1]);
    //drawSphere(1.5);  // now attached to robot 1
    glPopMatrix();
    glPushMatrix();    
    glColor3f(0.0,1.0,1.0); 
    glTranslatef(-HapticDevicePosition2[2],-HapticDevicePosition2[0]+SEPARATION,HapticDevicePosition2[1]);
    drawSphere(1.5);
    glPopMatrix();
  }
  glColor3f(1.0,1.0,1.0);
  glDisable(GL_COLOR_MATERIAL);
}

void drawPivot(float disp, float height, float width, float offset){
 drawCylinderOffset(disp,1.0,offset);
 glPushMatrix();
 glTranslatef(0.0,-disp,offset+height/2.0);
 glScalef(width,2.0,height);
 glutSolidCube(1.0);
 glPopMatrix();
 glPushMatrix();
 glTranslatef(0.0, disp,offset+height/2.0);
 glScalef(width,2.0,height);
 glutSolidCube(1.0);
 glPopMatrix();
}

void GetRobotTransform(GLfloat me[4][4], GLfloat mp[4][4], GLfloat mua[4][4], GLfloat mla[4][4]){
 int i;
 GLfloat cameraViewMatrix[16],m[16],pe[4],zero[4]={0.0,0.0,0.0,1.0};
 glGetFloatv(GL_MODELVIEW_MATRIX, cameraViewMatrix);
 glLoadIdentity();

 glRotatef(rs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 glTranslatef(0.0f,0.0f,baseHeight);
 glRotatef(rs2,0.0,1.0,0.0);   // second axis S2  around local y
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   
 m16to4(m,mua);  // pivot point (upperarm)

 glRotatef(rs3,0.0,0.0,1.0);   // 3rd axis S3
 glTranslatef(0.0f,0.0f,upperArm);
 glRotatef(re1,0.0,1.0,0.0);  // 4th axis E1
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   
 m16to4(m,mla);  // pivot point (lowerarm)

 glRotatef(re2,0.0,0.0,1.0);   // 5th axis E2
 glTranslatef(0.0f,0.0f,lowerArm);
 glRotatef(rw1,0.0,1.0,0.0);  // 6th axis W1
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   // to go from OGL order to C++ ordering
 m16to4(m,mp);  // pivot point (wrist)

 glRotatef(rw2,0.0,0.0,1.0);   // 7th axis W2
 glTranslatef(0.0f,0.0f,Hand);
 glGetFloatv(GL_MODELVIEW_MATRIX, m);
 TransposeMatrix16(m);   // to go from OGL order to C++ ordering
 m16to4(m,me);

 glMatrixMode(GL_MODELVIEW);
 glLoadMatrixf(cameraViewMatrix);
 get_transform=0;

 printf("Get Open GL Transform\n");
 printf("Angles:\n[S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E1=%5.1f] [W1=%5.1f W2=%5.1f]\n",
        rs1,rs2,rs3,re1,re2,rw1,rw2);
 rs1a[0]=rs1; rs2a[0]=rs2; rs3a[0]=rs3; re1a[0]=re1; re2a[0]=re2; rw1a[0]=rw1; rw2a[0]=rw2;
 printf("Transform Matrix End Point\n");
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[0][0],me[0][1],me[0][2],me[0][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[1][0],me[1][1],me[1][2],me[1][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[2][0],me[2][1],me[2][2],me[2][3]); 
 printf("%6.2f %6.2f %6.2f %6.2f\n",me[3][0],me[3][1],me[3][2],me[3][3]);

 m4by1(me,zero,pe);
 p_endpoint[0]=pe[0]; p_endpoint[1]=pe[1]; p_endpoint[2]=pe[2];
 printf("Graphic End point:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mp,zero,pe);
 p_wrist[0]=pe[0]; p_wrist[1]=pe[1]; p_wrist[2]=pe[2];
 for(i=0;i<3;i++)p_target[i]=p_wrist[i]; // set Target to current end point too!
 printf("Graphic Wrist    :(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mla,zero,pe);
 p_lowerarm[0]=pe[0]; p_lowerarm[1]=pe[1]; p_lowerarm[2]=pe[2];
 printf("Graphic Lower arm:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
 m4by1(mua,zero,pe);
 p_upperarm[0]=pe[0]; p_upperarm[1]=pe[1]; p_upperarm[2]=pe[2];
 printf("Graphic Upper arm:(xyz) (%6.2f  %6.2f %6.2f)\n", pe[0],pe[1],pe[2]);
}

void drawCCDAxes() {
 GLfloat m[16],m4[4][4];
 drawAxes(30.0);   // base axis 
 drawSphere(2.0); 
 glPushMatrix();
 m4by4(m_camera,B,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BU,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BUL,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 m4by4(m_camera,BULH,m4); m4to16(m4,m); TransposeMatrix16(m);  glLoadMatrixf(m);
 drawSphere(2.0);   drawAxes(5.0);
 glPopMatrix();
}

void drawRobotAxes() {
 drawAxes(30.0);   // base axis   [I]
 glPushMatrix();

 glRotatef(rs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 glTranslatef(0.0f,0.0f,baseHeight);
 glRotatef(rs2,0.0,1.0,0.0);   // second axis S2  around local y
 drawAxes(15.0);   // first joint

 glRotatef(rs3,0.0,0.0,1.0);   // 3rd axis S3
 glTranslatef(0.0f,0.0f,upperArm);
 glRotatef(re1,0.0,1.0,0.0);  // 4th axis E1
 drawAxes(10.0);  

 glRotatef(re2,0.0,0.0,1.0);   // 5th axis E2
 glTranslatef(0.0f,0.0f,lowerArm);
 glRotatef(rw1,0.0,1.0,0.0);  // 6th axis W1
 drawAxes(10.0);

 glRotatef(rw2,0.0,0.0,1.0);   // 7th axis W2

 glTranslatef(0.0f,0.0f,Hand);
 drawAxes(5.0);   // Final axis  (M)
 drawSphere(1.0); 

 glPopMatrix();
}

void drawRobotGeometry(int id) {
 float Lrs1,Lrs2,Lrs3,Lre1,Lre2,Lrw1,Lrw2;
 if(id == 1){Lrs1=rs1; Lrs2=rs2; Lrs3=rs3; Lre1=re1; Lre2=re2; Lrw1=rw1; Lrw2=rw2;}
 else       {Lrs1=qs1; Lrs2=qs2; Lrs3=qs3; Lre1=qe1; Lre2=qe2; Lrw1=qw1; Lrw2=qw2;}

 glPushMatrix();      // draw base disk and end point proxy calculated from OGL matrix
 glTranslatef(0.0,0.0,-0.5);
 drawDisk(75.0);
 glPopMatrix();       // end of base disk

 drawCylinder(15.0,15.0);
 glPushMatrix();
 glRotatef(Lrs1,0.0,0.0,1.0);  // first axis  S1  around local z  (applies to all children)
 drawPivot(8,baseHeight-15.0+5.0,10.0, 15.0);
 glTranslatef(0.0f,0.0f,baseHeight);
 drawCylinderH(5.0,15.0);
 drawCylinderH(1.0,20.0);
 glRotatef(Lrs2,0.0,1.0,0.0);   // second axis S2  around local y
 if(axesVis)drawWireCylinder(5,upperArm-15.0,0.0); else drawCylinder(5,upperArm-15.0);  

 glRotatef(Lrs3,0.0,0.0,1.0);   // 3rd axis S3
 drawPivot(8,15.0,10.0, upperArm-15.0);

 glTranslatef(0.0f,0.0f,upperArm);
 drawCylinderH(3.0,12.0);
 drawCylinderH(1.0,15.0);
 glRotatef(Lre1,0.0,1.0,0.0);  // 4th axis E1
 if(axesVis)drawWireCylinder(5,lowerArm-20.0,0.0);  else drawCylinder(5,lowerArm-20.0);

 glRotatef(Lre2,0.0,0.0,1.0);   // 5th axis E2
 drawPivot(6,20.0,6.0, lowerArm-20.0);
 
 glTranslatef(0.0f,0.0f,lowerArm);
 drawCylinderH(1.0,18.0);  // pin
 glRotatef(Lrw1,0.0,1.0,0.0);  // 6th axis W1
//drawWireCylinder(2.5,12.0,-12.0);
 drawCylinderOffset(2.5,12.0,-12.0);

 glRotatef(Lrw2,0.0,0.0,1.0);   // 7th axis W2
 if(axesVis)drawWireCylinder(3.0,8.0,0.0); else drawCylinder(3.0,8.0);  //The hand
 
 glTranslatef(0.0f,0.0f,Hand);
 if(!axesVis){
   drawSphere(2.0);
   drawWireCylinder(1.0,3.0,0.0);
 }
 glPopMatrix();
}

void display(void) {   // this is called at real time
  static GLfloat light0_ambient[] =  {0.1f, 0.1f, 0.3f, 1.0f};
  static GLfloat light0_diffuse[] =  {.8f, .8f, 1.0f, 1.0f};
  static GLfloat light0_position[] = {.5f, .5f, 1.0f, 0.0f};
  GLfloat m[16];
  GLdouble ex,ey,ez;
	glClearDepth(1.0f);
	glClearColor(0.1f, 0.1f, 0.2f, 0.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH);
  //glShadeModel(GL_FLAT);
  glEnable( GL_LIGHTING );
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse );
  glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient );

  if(bPhantoms > 0)UpdateFromPhantoms();
  glPushMatrix();
  glLoadIdentity();
  //  glLightfv(GL_LIGHT0, GL_POSITION, light0_position); // the light stays in this position relative to camera
  if(viewdir == 0){
    ez=er*cos(theta)*cos(phi);
    ex=er*cos(theta)*sin(phi);
    ey=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, SEPARATION/2, 50, 0, 1, 0);    // y - up
  }
  else if(viewdir == 1){
    ex=er*cos(theta)*cos(phi);
    ey=er*cos(theta)*sin(phi);
    ez=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, SEPARATION/2, 50, 0, 0, 1);    // z - up
  }
  else if(viewdir == 2){
    ey=er*cos(theta)*cos(phi);
    ez=er*cos(theta)*sin(phi);
    ex=er*sin(theta);
    gluLookAt(ex, ey, ez+50, 0, SEPARATION/2, 50, 1, 0, 0);    // x - up
  }
  glEnable( GL_NORMALIZE );

  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);  // light stays fixed relative to objects 

  ////////////// just to make sure our transform functions are OK  read the OGL convert and convert back
  glGetFloatv(GL_MODELVIEW_MATRIX, m);
  TransposeMatrix16(m);  // OGL to C matrix order  
  m16to4(m,m_camera);    // model view matrix of camera
  m4to16(m_camera,m);
  TransposeMatrix16(m);  // C to OGL matrix ordering  
  glMatrixMode(GL_MODELVIEW);
  glLoadMatrixf(m);
  ///////////////// 
  drawTargetAndProxy();
  if(modelVis){
    drawRobotGeometry(1);
    glPushMatrix();
    glTranslatef(0.0,SEPARATION,0.0);
    drawRobotGeometry(2);
    glPopMatrix();
  }
  if(axesVis)drawRobotAxes();
  if(ccdVis)drawCCDAxes();
  if(get_transform)GetRobotTransform(m_endpoint,m_wristpoint,m_upperarm,m_lowerarm);
  glPopMatrix();
  UpdateRobotServerData();
  glutSwapBuffers();
}

void reshape(GLint width, GLint height){
   glViewport(0, 0, width, height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   //gluPerspective(65.0, (float)width / height, 0.01, 1000);  // too close for Intel graphics to handle !! 
   gluPerspective(65.0, (float)width / height, 3, 800);
   glMatrixMode(GL_MODELVIEW);

   glLoadIdentity();
   if(bPhantoms > 0)UpdateHapticSpace();
}

// UI stuff ///////////////////////////////////////

static void UpdateFromPhantoms(void){
  UpdateHapticState();
  // need to set proxy position and orientations
  if(continuous){
    p_target[0] =-HapticDevicePosition1[2];
    p_target[1] =-HapticDevicePosition1[0];
    p_target[2] = HapticDevicePosition1[1];
    p_target2[0] = -HapticDevicePosition2[2];
    p_target2[1] = -HapticDevicePosition2[0];
    p_target2[2] =  HapticDevicePosition2[1];
    RunCCD(0,1);
    RunCCD(0,2);
  }
}

void Special(int key, int x, int y){
  int bchange=0,bShift=0;
  float step_size=0.1;
  float angl_step=0.5;
  unsigned long bmModifiers;
  bmModifiers = glutGetModifiers();
  if((bmModifiers & GLUT_ACTIVE_SHIFT) == GLUT_ACTIVE_SHIFT)bShift=1; else bShift=0;
  if((bmModifiers & GLUT_ACTIVE_CTRL) == GLUT_ACTIVE_CTRL)step_size=1.0; else step_size=0.1;
  if(key == GLUT_KEY_RIGHT){
     if(bShift)hand_pan += angl_step;
     else      p_target[1] += step_size; 
     bchange=1;
  } 
  if(key == GLUT_KEY_LEFT) {
     if(bShift)hand_pan -= angl_step;
     else     p_target[1] -= step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_UP)   {
     if(bShift)hand_tilt += angl_step;
     else     p_target[0] += step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_DOWN) {
     if(bShift)hand_tilt -= angl_step;
     else     p_target[0] -= step_size;  
     bchange=1;
  }  
  if(key == GLUT_KEY_PAGE_UP){
     if(bShift)hand_twist += angl_step;
     else     p_target[2] += step_size;  
     bchange=1;
  }
  if(key == GLUT_KEY_PAGE_DOWN){
     if(bShift)hand_twist -= angl_step;
     else     p_target[2] -= step_size;  
     bchange=1;
  }
  if(bchange && interactive){
    RunCCD(0,1);
    RunCCD(0,2);
  }
}

void Keyboard(unsigned char key, int x, int y){
  if(key==27) {
    if(bPhantoms > 0)exitPhantoms();
    exit(0);
  }
  if(key=='1')rs1 += 2.0;
  if(key=='2')rs1 -= 2.0;
  if(key=='3')rs2 += 2.0;
  if(key=='4')rs2 -= 2.0;
  if(key=='5')rs3 += 2.0;
  if(key=='6')rs3 -= 2.0;
  if(key=='7')re1 += 2.0;
  if(key=='8')re1 -= 2.0;
  if(key=='9')re2 += 2.0;
  if(key=='0')re2 -= 2.0;

  if(key=='-')rw1 += 2.0;  // hand end
  if(key=='=')rw1 -= 2.0;
  if(key=='q')rw2 += 2.0;  // hand twist angle 
  if(key=='w')rw2 -= 2.0;

  if(key=='!')qs1 += 2.0;
  if(key=='"')qs1 -= 2.0;
  if(key=='E')qs2 += 2.0;
  if(key=='$')qs2 -= 2.0;
  if(key=='%')qs3 += 2.0;
  if(key=='^')qs3 -= 2.0;
  if(key=='&')qe1 += 2.0;
  if(key=='*')qe1 -= 2.0;
  if(key=='(')qe2 += 2.0;
  if(key==')')qe2 -= 2.0;

  if(key=='_')qw1 += 2.0;  // hand end
  if(key=='+')qw1 -= 2.0;
  if(key=='Q')qw2 += 2.0;  // hand twist angle 
  if(key=='W')qw2 -= 2.0;



  if(key=='y')viewdir=0;
  if(key=='z')viewdir=1;
  if(key=='x')viewdir=2;
  if(key=='g')GetAnglesFromFile();
  if(key=='#'){
    printf("Angles: [S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E2=%5.1f] [W1=%5.1f W2=%5.1f]\n",
     rs1,rs2,rs3,re1,re2,rw1,rw2);
  }
  if(key=='.'){   // draw the target point directly
    get_transform=1;
  }
  if(key=='@'){
    RunCCD(1,1);
    RunCCD(1,2);
  }
}

static int bMouseDown=0,bMouseRightDown, bShiftModifier=0;
static int xd,yd,xrd,yrd;

void mouse(int button, int state, int x, int y){
   //#define GLUT_ACTIVE_CTRL                2
   //#define GLUT_ACTIVE_ALT                 4
   unsigned long bmModifiers;
   bmModifiers = glutGetModifiers();
   if((bmModifiers & GLUT_ACTIVE_SHIFT) == GLUT_ACTIVE_SHIFT)bShiftModifier=1; else bShiftModifier=0;
   if (button == GLUT_LEFT_BUTTON)   {
      if (state == GLUT_UP)      {
        bMouseDown=0;
      }
      else {
        xd=x; yd=y;
        bMouseDown=1;
      }
   }
   if (button == GLUT_RIGHT_BUTTON)   {
      if (state == GLUT_UP)      {
        bMouseRightDown=0;
      }
      else {
        xrd=x; yrd=y; 
        bMouseRightDown=1;
      }
   }
   if (button == GLUT_MIDDLE_BUTTON)   {
      if (state == GLUT_UP)      {
      }
      else {
      }
   }
}

void motion(int x, int y){
  GLdouble dx,dy,dr;
  if(bMouseDown == 1){
    dx=x-xd; dy=y-yd;
    if(bShiftModifier){
      phi   -= dx*0.03;
      theta += dy*0.04;
    }
    else{
      phi   -= dx*0.003;
      theta += dy*0.004;
    }
    xd=x; yd=y;
  }
  if(bMouseRightDown == 1){
    dr=x-xrd;
    er -= dr*0.05;
    xrd=x;
  }
}

void menu(int value){
  switch (value) {
    case 100:
      modelVis ^= 1;
      glutPostRedisplay();
      break;
    case 104:
      axesVis ^= 1;
      glutPostRedisplay();
      break;
    case 114:
      ccdVis ^= 1;
      glutPostRedisplay();
      break;
    case 101:
      viewdir=0;
      glutPostRedisplay();
      break;
    case 102:
      viewdir=1;
      glutPostRedisplay();
      break;
    case 103:
      viewdir=2;
      glutPostRedisplay();
      break;
    case 109:
      printf("Angles: [S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E1=%5.1f] [W1=%5.1f W2=%5.1f]\n",
            rs1,rs2,rs3,re1,re2,rw1,rw2);
      break;
    case 111:
      get_transform=1;
      break;
    case 112:
      rs1=rs1a[0]; rs2=rs2a[0]; rs3=rs3a[0]; re1=re1a[0]; re2=re2a[0];  rw1=rw1a[0];  rw2=rw2a[0];
      glutPostRedisplay();
      break;
    case 110:
      rs1=0.0,rs2=0.0,rs3=0.0,re1=0.0,re2=0.0,rw1=0.0,rw2=0.0;
      p_endpoint[0]=20.0; p_endpoint[1]=0.0; p_endpoint[2]=0.0;
      hand_pan=hand_tilt=0.0; hand_twist=0.0;
      glutPostRedisplay();
      break;
    case 115:
      RunCCD(1,1);
      RunCCD(1,2);
      break;
    case 116:
      interactive ^=1;
      break;
    case 117:
      GetAnglesFromFile();
      break;
    case 118:
      continuous ^=1;
      break;
    case 666:
      if(bPhantoms > 0)exitPhantoms();
      exit(0);
      break;
    default:
      break;
  }
}

static void GetAnglesFromFile(void){
  FILE *fp;
  if((fp=fopen("pa10sim.txt","r")) != NULL){
    // s1 s2 s3  e1 e2  w1 w2
    fscanf(fp,"%f %f %f %f %f %f %f",&rs1,&rs2,&rs3,&re1,&re2,&rw1,&rw2); 
    fclose(fp);
  }
  else printf("No [pa10sim.txt] file\n");
}

static void AddMenuItems(void){
  int shapeMenu, viewMenu, actionMenu, renderModeMenu;
  shapeMenu = glutCreateMenu(menu);
  glutAddMenuEntry("Model", 100);
  glutAddMenuEntry("Axes", 104);
  glutAddMenuEntry("CCD Axes", 114);
  viewMenu = glutCreateMenu(menu);
  glutAddMenuEntry("Y - up ", 101);
  glutAddMenuEntry("Z - up ", 102);
  glutAddMenuEntry("X - up ", 103);
  renderModeMenu = glutCreateMenu(menu);
  glutAddMenuEntry("Last Proxy Pos", 112);
  actionMenu = glutCreateMenu(menu);
  glutAddMenuEntry("Position at Target (@)",115);
  glutAddMenuEntry("Print Angles (#)", 109);
  glutAddMenuEntry("Print Matrices / Move End Proxy (.)", 111);
  glutAddMenuEntry("Read Angles from [pat10sim.txt] (g)",117);
  glutCreateMenu(menu);
  glutAddSubMenu("Hide/Show", shapeMenu);
  glutAddSubMenu("View Direction  ", viewMenu);
  glutAddSubMenu("GoTo", renderModeMenu);
  glutAddSubMenu("Action", actionMenu);
  glutAddMenuEntry("Reset",110);
  glutAddMenuEntry("Toggle Interactive Target Tracking",116);
  glutAddMenuEntry("Toggle Continuous Tracking (Phantom)",118);
  glutAddMenuEntry("Quit", 666);
  //glutAttachMenu(GLUT_RIGHT_BUTTON);
  glutAttachMenu(GLUT_MIDDLE_BUTTON);
}

///////////  end UI stuff

static void CalculateCcdTransforms(void){
 float zero[4]={0,0,0,1};
 float m1[4][4];
 float s1[4][4],s2[4][4],s3[4][4],e1[4][4],e2[4][4],w1[4][4],w2[4][4],tb[4][4],tua[4][4],tla[4][4],th[4][4];
 rotz(s1,ccd_rs1);
 roty(s2,ccd_rs2);
 tram(tb,0.0,0.0,baseHeight);
 m4by4(tb,s2,m1);
 m4by4(s1,m1,B);  // base transformation
 rotz(s3,ccd_rs3);
 roty(e1,ccd_re1);
 tram(tua,0.0,0.0,upperArm);
 m4by4(tua,e1,m1);
 m4by4(s3,m1,U);  // upperarm transform - relative to base
 rotz(e2,ccd_re2);
 roty(w1,ccd_rw1);
 tram(tla,0.0,0.0,lowerArm);
 m4by4(tla,w1,m1);
 m4by4(e2,m1,L); // lowerarm - relaive to upperarm
 rotz(w2,ccd_rw2);
 tram(th,0.0,0.0,Hand);
 m4by4(w2,th,H); // Hand relative to lowerarm 
 m4by4(B,U,BU);   // composites
 m4by4(BU,L,BUL);  
 m4by4(BUL,H,BULH);
 m4by1(B,zero,c_upperarm);  //at top of base
 m4by1(BU,zero,c_lowerarm);
 m4by1(BUL,zero,c_wrist);   // this is the primary pivot point
 m4by1(BULH,zero,c_hand);   // end point
}

static void RotateVector(float ct, float st, float r[4], float n[4]){
 int i;
 float t1[4],t2[4],t3[4],t4[4],cc,ndr;
 VECSCALE4(ct,r,t1)
 ndr=DOT4(n,r);
 cc=(1.0-ct)*ndr;
 VECSCALE4(cc,n,t2)
 CROSS4(n,r,t3)
 VECSCALE(st,t3,t4)
 for(i=0;i<4;i++){
   r[i]=t1[i]+t2[i]+t4[i];
 }
}

static int TargetReached(float *e, float p_t[]){
 float d,errorD=0.1,dend[4];
 VECSUB4(p_t,c_wrist,dend)
 d=sqrt(dend[0]*dend[0]+dend[1]*dend[1]+dend[2]*dend[2]);
 *e=d;
 if(d < errorD)return 1;
 return 0;
}

static void SetHandDirection(int idd){
 float d;
 float ddv1[4],ddv2[4],up[4]={0,0,1,1},dirn[4]={1,0,0,1};
 float mmr[4][4],mmi[4][4],hh[4][4];
 int i,j;

 // this will not be used if Phantoms active
 if(bPhantoms == 0){
   float r1[4][4],r2[4][4],r3[4][4],ti[4][4];
   dirn[0]=cos(hand_pan*DEGRAD)*cos(hand_tilt*DEGRAD);
   dirn[1]=sin(hand_pan*DEGRAD)*cos(hand_tilt*DEGRAD);
   dirn[2]=sin(hand_tilt*DEGRAD);
   normalise(dirn);
   //rotz(r1,hand_twist+180.0);         // make transform for end of tool effector
   rotz(r1,hand_twist);         // make transform for end of tool effector
   roty(r2,90.0f-hand_tilt);   // to send to robot
   rotz(r3,hand_pan);
   m4by4(r2,r1,ti);
   if(idd == 1)m4by4(r3,ti,ToolO);
   else        m4by4(r3,ti,ToolO2);
 }
 else{
   float r1,r2,r3,dirn1[4]={0,1,0,1},r180[4][4],up1[4]={1,0,0,1};
   float mt1[4][4],mt2[4][4],mt3[4][4],mt4[4][4],mt5[4][4];
   for(i=0;i<4;i++)for(j=0;j<4;j++){
     if(idd == 1)hh[i][j]=HapticDevRotationMatrix1[i][j]; 
     else        hh[i][j]=HapticDevRotationMatrix2[i][j]; 
   }
   {
     EulerAngles outAng;
     outAng=Eul_FromHMatrix(hh,EulOrdXYZr);
     //     roty(mt1,(angx = -outAng.x*RADDEG*0.5)+90);
     //     rotz(mt2,(angy =  outAng.y*RADDEG*0.5));
     roty(mt1,(angx = -outAng.x*RADDEG*0.9)+90);
     rotz(mt2,(angy =  outAng.y*RADDEG*0.8));
     //     rotz(mt3,(angz = -outAng.z*RADDEG*0.5)+180.0);  // to rotate the tool end
     rotz(mt3,(angz = -outAng.z*RADDEG*0.5));
     if(idd == 1){
       null_transform(ToolO);
       m4by4(mt3,ToolO,mt4);
       m4by4(mt1,mt4,mt5);
       m4by4(mt2,mt5,ToolO);
     }
     else{
       null_transform(ToolO2);
       m4by4(mt3,ToolO2,mt4);
       m4by4(mt1,mt4,mt5);
       m4by4(mt2,mt5,ToolO2);
     }
   }
   ConvertHLtoLocal(hh,mmr);
   rotz(r180,180.0);
   m4by4(mmr,r180,hh);
   m4by1(hh,dirn1,dirn);
   m4by1(hh,up1,up);
 }
 VECSUM4(c_wrist,TOFFSET*dirn,c_target)  // indicate the direction "dirn" from wrist
 CalculateCcdTransforms();
 ExtractMatrixRotation(BU,mmr);   
 InvertMatrixRotation(mmr,mmi,1);
 m4by1(mmi,dirn,ddv2);                    // direction is now relative to GLOBAL axes
 ccd_re2=atan2(ddv2[1],ddv2[0])*RADDEG;
 d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
 ccd_rw1=90.0-atan2(ddv2[2],d)*RADDEG;

 CalculateCcdTransforms();
 ExtractMatrixRotation(BUL,mmr);          // turn the hand so that its "X" axis is vertical
 InvertMatrixRotation(mmr,mmi,1); 
 m4by1(mmi,up,ddv2);
 ccd_rw2=atan2(ddv2[1],ddv2[0])*RADDEG + hand_twist;

 CalculateCcdTransforms();
 if(idd == 1){
   re2=ccd_re2, rw1=ccd_rw1;  rw2=ccd_rw2;  // send the angles to the robot
   VECCOPY(c_hand,p_hand)   // end point

 }
 else{
   qe2=ccd_re2, qw1=ccd_rw1;  qw2=ccd_rw2;  // send the angles to the robot
   VECCOPY(c_hand,p_hand2)   // end point
 }
}

static void RunCCD(int prnt, int ID){
 int i,nsteps,stepMax=200;
 float d,ct,st,error;
 float nr[4],ddv1[4],ddv2[4];
 float mmr[4][4],mmi[4][4];
 float p_tar[3];

 // starting angles for CCD
 ccd_rs1=0.0; ccd_rs2=0.0; // upper arm angles
 ccd_rs3=0.0; ccd_re1=0.0; // lower arm angles
 ccd_re2=0.0; ccd_rw1=0.0; ccd_rw2=0.0;  // wrist hand angles (determined later)
 nsteps=0;
 if(ID == 1){for(i=0;i<3;i++)p_tar[i]=p_target[i];}
 if(ID == 2){for(i=0;i<3;i++)p_tar[i]=p_target2[i];}

 LOOP:   // Simple 2 stage CCD 
 CalculateCcdTransforms();
 if(ccdRLT && prnt){ // for rteal-time display
    if(ID == 2){qs1=ccd_rs1; qs2=ccd_rs2;  qs3=ccd_rs3; qe1=ccd_re1; }
    else       {rs1=ccd_rs1; rs2=ccd_rs2;  rs3=ccd_rs3; re1=ccd_re1; }
    display();
    Sleep(250);
 } 
 if(TargetReached(&error,p_tar))goto SOLVED;
 if(nsteps > stepMax)goto FAILED;
 // pivot lower arm - to align lower arm in direction of target 
 // (and place c_wrist on line joining c_lowerarm with p_tar)
 VECSUB4(p_tar,c_lowerarm,ddv1)
 ExtractMatrixRotation(B,mmr);   
 InvertMatrixRotation(mmr,mmi,1); 
 m4by1(mmi,ddv1,ddv2);   
 ccd_rs3=atan2(ddv2[1],ddv2[0])*RADDEG;
 d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
 ccd_re1=90.0-atan2(ddv2[2],d)*RADDEG;
 CalculateCcdTransforms();
 if(TargetReached(&error,p_tar))goto SOLVED;
 // pivot upper arm - to place c_wrist on a line joining c_upperarm with p_tar
 // this step would need to be repeated if there were more than two links in the chain
 VECSUB(c_wrist ,c_upperarm,ddv1)  normalise(ddv1);
 VECSUB(p_tar,   c_upperarm,ddv2)  normalise(ddv2);
 if(fabs(DOT4(ddv1,ddv2)) > 0.0001){ // not on line so need to rotate to it
   CROSS4(ddv1,ddv2,nr) normalise(nr);
   ct=DOT4(ddv1,ddv2); st=sqrt(1-ct*ct);
   VECSUB4(c_lowerarm,c_upperarm,ddv1)
   RotateVector(ct,st,ddv1,nr);
   ExtractMatrixRotation(II,mmr);  // This is wasteful since CCD has only two links
   InvertMatrixRotation(mmr,mmi,1); 
   m4by1(mmi,ddv1,ddv2);   
   ccd_rs1=atan2(ddv2[1],ddv2[0])*RADDEG;
   d=sqrt(ddv2[0]*ddv2[0]+ddv2[1]*ddv2[1]);
   ccd_rs2=90.0-atan2(ddv2[2],d)*RADDEG;
 } 
 nsteps++;
 //printf("Error %7.3f\n",error);
 goto LOOP; 
 FAILED: ;  // this most likely occurs when target it out of range
 SOLVED: 
 // End of CCD algorithm - assign to robot
 
 if(ID == 2){
   qs1=ccd_rs1; qs2=ccd_rs2;     qs3=ccd_rs3; qe1=ccd_re1; 
   qe2=0.0; qw1=0.0; qw2=0.0;  
 }
 else{
   rs1=ccd_rs1; rs2=ccd_rs2;     rs3=ccd_rs3; re1=ccd_re1; 
   re2=0.0; rw1=0.0; rw2=0.0;  // will be calculated to keep the robot's hand pointing in the direction "dirn"
 }

 // now calculate the direction of the hand to keep it pointing in the desired direction
 SetHandDirection(ID);

 if(prnt){
  printf("CCD Algorithm - Solved in %d iterations (Error %7.3f)\n",nsteps,error);
  printf("TARGET:(xyz)         (%6.2f  %6.2f %6.2f)\n", p_tar[0],p_tar[1],p_tar[2]);
  printf("CCD Positions\n");
  printf("Hand      base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_hand[0],c_hand[1],c_hand[2]);
  printf("Wrist     base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_wrist[0],c_wrist[1],c_wrist[2]);
  printf("Lower arm base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_lowerarm[0],c_lowerarm[1],c_lowerarm[2]);
  printf("Upper arm base:(xyz) (%6.2f  %6.2f %6.2f)\n", c_upperarm[0],c_upperarm[1],c_upperarm[2]);
  printf("CCD Angles\n[S1=%5.1f S2=%5.1f S3=%5.1f] [E1=%5.1f E2=%5.1f] [W1=%5.1f W2=%5.1f]\n",
      ccd_rs1,ccd_rs2,ccd_rs3,ccd_re1,ccd_re2,ccd_rw1,ccd_rw2);
  printf("CCD EndPoint Transform\n");
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[0][0],BULH[0][1],BULH[0][2],BULH[0][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[1][0],BULH[1][1],BULH[1][2],BULH[1][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[2][0],BULH[2][1],BULH[2][2],BULH[2][3]); 
  printf("%6.2f %6.2f %6.2f %6.2f\n",BULH[3][0],BULH[3][1],BULH[3][2],BULH[3][3]); 

  {float h[4],X[4]={1,0,0,1},Y[4]={0,1,0,1},Z[4]={0,0,1,1};
   ExtractMatrixRotation(BULH,mmr);   
   m4by1(mmr,X,h);  printf("Hand Axes     : X=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
   m4by1(mmr,Y,h);  printf("              : Y=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
   m4by1(mmr,Z,h);  printf("              : Z=(%6.2f %6.2f %6.2f)\n",h[0],h[1],h[2]);
  }
  if(bPhantoms > 0){
    printf("Phantom Position 1 (%6.2f  %6.2f %6.2f)\n", 
      HapticDevicePosition1[0],HapticDevicePosition1[1],HapticDevicePosition1[2]);
    printf("Phantom          2 (%6.2f  %6.2f %6.2f)\n", 
      HapticDevicePosition2[0],HapticDevicePosition2[1],HapticDevicePosition2[2]);
  }
 }
}

int main(int argc, char* argv[]) {
  int window_id;
  glutInit(&argc, argv);
  glutInitWindowPosition(400,300);
  glutInitWindowSize(800,600);
  //glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  window_id = glutCreateWindow( "Dual PA10 Robot Viewer" );
  glutSetWindow( window_id );
  glutDisplayFunc ( display  );
  glutReshapeFunc (reshape);
  glutKeyboardFunc(Keyboard);
  glutSpecialFunc(Special);
  glutIdleFunc( display );
  glutMotionFunc(motion);
  glutMouseFunc(mouse);
  AddMenuItems();
  bPhantoms=initPhantoms();
  glutMainLoop();
  printf("Program Ended\n");
}

//////////////////  Geometry & Utility Functions belowe //////////
//////////////////////////////////////////////////////////////////

// This also sets the Euler angle rep of rotation matrix and verifys that the inverse gives the same matrix
static void verifyEulerConversion(HMatrix h1, HMatrix h2){
 int error=0;
 float eps=0.01;
 int i,j;
 HMatrix h;
 Robo1_E=Eul_FromHMatrix(h1,EulOrdZYZr);
 Robo1_E.w=EulOrdZYZr;
 Eul_ToHMatrix(Robo1_E,h);
 error=0;
 for(i=0;i<3;i++)for(j=0;j<3;j++){
  if(fabs(h1[i][j] - h[i][j]) > eps)error=1;
 }
 if(error){
   printf("Matrix 1 differs\n");
   for(i=0;i<3;i++){
     printf("%6.2f %6.2f %6.2f  |  %6.2f %6.2f %6.2f\n", h1[i][0],h1[i][1],h1[i][2],h[i][0],h[i][1],h[i][2]);
   }
 } 
 Robo2_E=Eul_FromHMatrix(h2,EulOrdZYZr);
 Robo2_E.w=EulOrdZYZr;
 Eul_ToHMatrix(Robo2_E,h);
 error=0;
 for(i=0;i<3;i++)for(j=0;j<3;j++){
  if(fabs(h2[i][j] - h[i][j]) > eps)error=1;
 }
 if(error){
   printf("Matrix 2 differs\n");
   for(i=0;i<3;i++){
     printf("%6.2f %6.2f %6.2f  |  %6.2f %6.2f %6.2f\n", h2[i][0],h2[i][1],h2[i][2],h[i][0],h[i][1],h[i][2]);
   }
 } 
}

// use of this SHOULD BE REPLACE BY funciton that obtains the rotation inverse by using
// THE TRANSPOSE since it is the SAME the full inverse for ROTATIONs
static void InvertMatrixRotation(float T[4][4], float t[4][4], int zero_translation){
 int i,j;
 float a[3][3];
 float det;
 null_transform(t);
 if(zero_translation){for(i=0;i<3;i++)for(j=0;j<3;j++)a[i][j]=T[i][j];}
 else                {for(i=0;i<4;i++)for(j=0;j<4;j++)a[i][j]=T[i][j];}
 det=a[0][0]*(a[1][1]*a[2][2] - a[2][1]*a[1][2])
    -a[0][1]*(a[1][0]*a[2][2] - a[2][0]*a[1][2])
    +a[0][2]*(a[1][0]*a[2][1] - a[2][0]*a[1][1]);
 if(fabs(det) < 1.e-10)return;
 det = 1.0/det;
 t[0][0] = +(a[1][1]*a[2][2] - a[2][1]*a[1][2])*det;
 t[1][0] = -(a[1][0]*a[2][2] - a[2][0]*a[1][2])*det;
 t[2][0] = +(a[1][0]*a[2][1] - a[2][0]*a[1][1])*det;
 t[0][1] = -(a[0][1]*a[2][2] - a[2][1]*a[0][2])*det;
 t[1][1] = +(a[0][0]*a[2][2] - a[2][0]*a[0][2])*det;
 t[2][1] = -(a[0][0]*a[2][1] - a[2][0]*a[0][1])*det;
 t[0][2] = +(a[0][1]*a[1][2] - a[1][1]*a[0][2])*det;
 t[1][2] = -(a[0][0]*a[1][2] - a[1][0]*a[0][2])*det;
 t[2][2] = +(a[0][0]*a[1][1] - a[1][0]*a[0][1])*det;
}

static void InvertMatrixRotationT(float T[4][4], float t[4][4], int zero_rotation){ // using transpose
 int i,j;
 null_transform(t);
 if(!zero_rotation){for(i=0;i<4;i++){t[i][3]=T[i][3]; t[3][i]=T[3][i];  } }
 for(i=0;i<3;i++) for(j=0;j<3;j++) {
   t[i][j]=T[j][i];
 }
}

static void ExtractMatrixRotation(float T[4][4], float t[4][4]){  // get the rotation part of a matrix
 int i,j;
 float det;
 null_transform(t);
 for(i=0;i<3;i++)for(j=0;j<3;j++)t[i][j]=T[i][j];
}
 
static void SetMatrixRotation(float T[4][4], float r[4][4], float t[4][4]){ // set the rotation part of matrix
 int i,j;
 float det;
 for(i=0;i<4;i++)for(j=0;j<4;j++)t[i][j]=T[i][j];
 for(i=0;i<3;i++)for(j=0;j<3;j++)t[i][j]=r[i][j];
}


int InvertMatrix(float *A, long n, float *B){
 int i,j,k,imax,kp1;
 float del,amax,atmp,btmp,div,amult,eps=1.e-6;
 for(i=0;i<n;i++)for(j=0;j<n;j++){
   if(i == j)B[i*n+j]=1.0;
   else      B[i*n+j]=0.0;
 }
 del=1.0;
 for(k=0;k<n;k++){
   if(k < n-1){
     imax=k;
     amax=fabs(A[k*n+k]);
     kp1=k+1;
     for(i=kp1;i<n;i++){
       if(amax-fabs(A[i*n+k]) < 0.0){
         imax=i;
         amax=fabs(A[i*n+k]);
       }
     }
     if(imax != k){
       for(j=0;j<n;j++){
         atmp=A[imax*n+j];
         A[imax*n+j]=A[k*n+j];
         A[k*n+j]=atmp;
         btmp=B[imax*n+j];
         B[imax*n+j]=B[k*n+j];
         B[k*n+j]=btmp;
       }
       del = -del;
     }
   }
   if(fabs(A[k*n+k]) < eps)return FALSE;
   del=A[k*n+k]*del;
   div=1.0/A[k*n+k];
   for(j=0;j<n;j++){
     A[k*n+j]=A[k*n+j]*div;
     B[k*n+j]=B[k*n+j]*div;
   }
   for(i=0;i<n;i++){
     amult=A[i*n+k];
     if(i != k){
       for(j=0;j<n;j++){
         A[i*n+j]=A[i*n+j]-amult*A[k*n+j];
         B[i*n+j]=B[i*n+j]-amult*B[k*n+j];
       }
     }
   }
 }
 return TRUE;
}

static void TransposeMatrix16(GLfloat m[16]){
  GLfloat t;
  t=m[ 4]; m[ 4]=m[ 1]; m[ 1]=t;
  t=m[ 6]; m[ 6]=m[ 9]; m[ 9]=t;
  t=m[11]; m[11]=m[14]; m[14]=t;
  t=m[ 2]; m[ 2]=m[ 8]; m[ 8]=t;
  t=m[ 3]; m[ 3]=m[12]; m[12]=t;
  t=m[ 7]; m[ 7]=m[13]; m[13]=t;
}


static void ConvertHLtoLocal(float mgl[4][4], float mfx[4][4]){
 float mt[4][4],temp[4][4];
 mt[0][0] = 0.0;  mt[0][1] = 0.0;   mt[0][2] =-1.0;   mt[0][3] = 0.0;
 mt[1][0] =-1.0;  mt[1][1] = 0.0;   mt[1][2] = 0.0;   mt[1][3] = 0.0;
 mt[2][0] = 0.0;  mt[2][1] = 1.0;   mt[2][2] = 0.0;   mt[2][3] = 0.0;
 mt[3][0] = 0.0;  mt[3][1] = 0.0;   mt[3][2] = 0.0;   mt[3][3] = 1.0;
 m4by4(mgl,mt,temp);
 m4by4(mt,temp,mfx);
 return;
}

static void m4by4(GLfloat t1[4][4], GLfloat t2[4][4], GLfloat tr[4][4]){   //tr = t1 * t2
  // tr = t1 X t2
 long i,j,k;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;

  for(k=0;k<4;k++)tr[i][j]=tr[i][j]+t1[i][k]*t2[k][j];
 }
 return;
}

static void m4by1(GLfloat t4[4][4], GLfloat m[4],  GLfloat r[4]){  // C++ ordering
   r[0]=t4[0][0]*m[0]+t4[0][1]*m[1]+t4[0][2]*m[2]+t4[0][3]*m[3];
   r[1]=t4[1][0]*m[0]+t4[1][1]*m[1]+t4[1][2]*m[2]+t4[1][3]*m[3];
   r[2]=t4[2][0]*m[0]+t4[2][1]*m[1]+t4[2][2]*m[2]+t4[2][3]*m[3];
   r[3]=t4[3][0]*m[0]+t4[3][1]*m[1]+t4[3][2]*m[2]+t4[3][3]*m[3];
}

static void m16to4(GLfloat m[16], GLfloat v[4][4]){  // C++ ordering
  v[0][0]=m[ 0]; v[0][1]=m[ 1]; v[0][2]=m[ 2]; v[0][3]=m[ 3];
  v[1][0]=m[ 4]; v[1][1]=m[ 5]; v[1][2]=m[ 6]; v[1][3]=m[ 7];
  v[2][0]=m[ 8]; v[2][1]=m[ 9]; v[2][2]=m[10]; v[2][3]=m[11];
  v[3][0]=m[12]; v[3][1]=m[13]; v[3][2]=m[14]; v[3][3]=m[15];
}

static void m4to16(GLfloat v[4][4], GLfloat m[16]){   // C++ ordering
  m[ 0]= v[0][0]; m[ 1]=v[0][1]; m[ 2]=v[0][2]; m[ 3]=v[0][3];
  m[ 4]= v[1][0]; m[ 5]=v[1][1]; m[ 6]=v[1][2]; m[ 7]=v[1][3];
  m[ 8]= v[2][0]; m[ 9]=v[2][1]; m[10]=v[2][2]; m[11]=v[2][3];
  m[12]= v[3][0]; m[13]=v[3][1]; m[14]=v[3][2]; m[15]=v[3][3];
}

static void rotz(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][1] = -sin(ang);
  tr[1][0] =  sin(ang);
  tr[1][1] =  cos(ang);
  return;
}

static void rotx(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[1][1] =  cos(ang);
  tr[1][2] = -sin(ang);
  tr[2][1] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

static void roty(float tr[4][4], float ang){
 long i,j;
 ang *= DEGRAD * (-1.0);   
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  tr[i][j]=0.0;
  if(i == j)tr[i][j]=1.0;
 }
  tr[0][0] =  cos(ang);
  tr[0][2] = -sin(ang);
  tr[2][0] =  sin(ang);
  tr[2][2] =  cos(ang);
  return;
}

static void tram(float t[4][4], float dx, float dy, float dz){
 long i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 {
  t[i][j]=0.0;
  if(i == j)t[i][j]=1.0;
 }
 t[0][3]=dx;
 t[1][3]=dy;
 t[2][3]=dz;
 t[3][3]=1;
 return;
}

static void c4to4(float tin[4][4], float tout[4][4]){
 int i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 tout[i][j]=tin[i][j];
 return;
}

static void null_transform(float t[4][4]){
 short i,j;
 for(i=0;i<4;i++)
 for(j=0;j<4;j++)
 if(i == j)t[i][j]=1.0;
 else      t[i][j]=0.0;
}


static void normalise(float v[4]){
 float n,nn;
 n= (float)((v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]));
 if(n < 1.e-15)return;
 nn=sqrt(n);
 n = 1.0 / nn;
 VECSCALE4(n,v,v)
 return;
}


//     C order
//
//     0 1 2 3 4 5 ....  Linear (vector) form
//
//     0   1   2   3     [0][0]  [0][1]  [0][2]  [0][3]  
//     4   5   6   7     [1][0]  [1][1]  [1][2]  [1][3]
//     8   9  10  11     [2][0]  [2][1]
//    12  13  14  15     [3][0]  [3][1]  

//  OGL ordering is TRANSPOSED when copied from vector (linear) form
